@@ -4,6 +4,7 @@ Downloads files and stores them locally.
 import re
 import os
 import zipfile
+from pathlib import Path
 
 from xbrl.helper.connection_manager import ConnectionManager
 
@@ -26,7 +27,7 @@ class HttpCache:
 
     """
 
-    def __init__(self, cache_dir: str, delay: int = 500):
+    def __init__(self, cache_dir: str, delay: int = 500, verify_https: bool = True):
         """
         :param cache_dir: Root directory of the disk cache (all requested files will be cached in this directory)
         :param delay: How many milliseconds should the cache wait, before requesting another file from the same server
@@ -34,9 +35,8 @@ class HttpCache:
         # check if the cache_dir ends with a /
         if not cache_dir.endswith('/'): cache_dir += '/'
         self.cache_dir: str = cache_dir
-        self.delay: int = delay
         self.headers: dict or None = None
-        self.connection_manager = ConnectionManager(delay)
+        self.connection_manager = ConnectionManager(delay, verify_https=verify_https)
 
     def set_headers(self, headers: dict) -> None:
         """
@@ -52,16 +52,16 @@ class HttpCache:
                               logs: bool = True) -> None:
         """
         Sets the connection params for all following request
-        :param delay: int specifying milliseconds to wait between each successfull request
+        :param delay: int specifying milliseconds to wait between each successful request
         :param retries: int specifying how many times a request will be tried before assuming its failure.
         :param backoff_factor: Used to measure time to sleep between failed requests. The formula used is:
             {backoff factor} * (2 ** ({number of total retries} - 1))
+        :param logs: enables or disables download logs
         :return:
         """
-        self.connection_manager._delay = delay
+        self.connection_manager._delay_ms = delay
         self.connection_manager._retries = retries
         self.connection_manager._backoff_factor = backoff_factor
-        self.connection_manager._delay = delay
         self.connection_manager.logs = logs
 
     def cache_file(self, file_url: str) -> str:
@@ -119,7 +119,7 @@ class HttpCache:
         """
         return self.cache_dir + re.sub("https?://", "", url)
 
-    def cache_edgar_enclosure(self, enclosure_url: str) -> None:
+    def cache_edgar_enclosure(self, enclosure_url: str) -> str:
         """
         The SEC provides zip folders that contain all xbrl related files for a given submission.
         These files are i.e: Instance Document, Extension Taxonomy, Linkbases.
@@ -130,7 +130,7 @@ class HttpCache:
         One way to get the zip enclosure url is through the Structured Disclosure RSS Feeds provided by the SEC:
         https://www.sec.gov/structureddata/rss-feeds-submitted-filings
         :param enclosure_url: url to the zip folder.
-        :return:
+        :return: relative path to extracted zip's content
         """
         if not enclosure_url.endswith('.zip'):
             raise Exception("This is not a valid zip folder")
@@ -141,3 +141,46 @@ class HttpCache:
         with zipfile.ZipFile(enclosure_path, "r") as zip_ref:
             zip_ref.extractall(submission_dir_path)
             zip_ref.close()
+        return submission_dir_path
+
+    def find_entry_file(self, dir_path: str) -> str or None:
+        """
+        NOTE: This function only works for enclosed SEC submissions that where already downloaded!
+        Also this function does only return the most likely file path for the instance document.
+        If you want to be certain i would recommend to use the SEC Structured Disclosure RSS Feeds
+        https://www.sec.gov/structureddata/rss-feeds-submitted-filings
+        These rss feeds list all files per submission and gives you information about the filetype (instance document,
+        taxonomy schema, label linkbase, exhibit e.t.c)
+
+        Find the most likely entry file in provided filling directory
+        """
+
+        # filter for files in interest
+        valid_files = []
+        for ext in '.htm .xml .xsd'.split():  # valid extensions in priority
+            for f in os.listdir(dir_path):
+                f_full = os.path.join(dir_path, f)
+                if os.path.isfile(f_full) and f.lower().endswith(ext):
+                    valid_files.append(f_full)
+
+        # find first file which is not included by others
+        entry_candidates = []
+        for file1 in valid_files:
+            f_dir, file_nm = os.path.split(file1)
+            # foreach file check all other for inclusion
+            found_in_other = False
+            for file2 in valid_files:
+                if file1 != file2:
+                    if file_nm in Path(file2).read_text():
+                        found_in_other = True
+                        break
+
+            if not found_in_other:
+                entry_candidates.append((file1, os.path.getsize(file1)))
+
+        # if multiple choose biggest
+        entry_candidates.sort(key=lambda tup: tup[1], reverse=True)
+        if len(entry_candidates) > 0:
+            file_path, size = entry_candidates[0]
+            return file_path
+        return None
